import csv
import json

class DataChunk:
    """
    Class to represent quired data

    Attributes:
        data (dict) : Dataframe like dictionary
    """

    def __init__(self, data: dict):
        """
        The constructor of DataChunk class

        Attributes:
            data (dict) : dataframe like dict
        """
        self.__data = data
        self.dict = data
        
    def saveas_csv(self, filepath: str):
        """
        Save DataChunk as CSV
        
        Parameters:
            filepath (str) : saving path
        """
        header = list(self.__data.keys())

        with open(filepath, 'w', encoding='UTF8') as f:
            writer = csv.writer(f)

            writer.writerow(header)

            for i in range(len(self.__data[header[0]])):
                writer.writerow([self.__data[d][i] for d in header])


    def saveas_json(self, filepath: str):
        """
        Save DataChunk as CSV
        
        Parameters:
            filepath (str) : saving path
        """
        with open(filepath, "w") as jsonf:
            json.dump(self.__data, jsonf)

    



import sqlite3

from corgidb.objects.table import Table
from corgidb.constants import *
from corgidb.exceptions import *

class Utils:
    """
    This is a class for executing basic sqlite3 operations or to be called utilities class

    Attributes:
        connection (sqlite3.Connection): a Connection to the database it self
    """

    def __init__(self, connection: sqlite3.Connection):
        """
        The constructor for Utils class

        Parameters:
            connection (sqlite3.Connection): a Connection to the database it self
        """
        self.connection = connection
        self.cursor     = self.connection.cursor()
        



    def sqlcmd(self, command: str):
        """
        This function to execute raw SQL command not recommened to be called if not necessary cannot return any data

        Parameters:
            command (str)   : Raw SQLite command
        
        Returns:
            list            : a list of fetched sql data return empty list if not fetched
        """
        with self.connection:
            if "SELECT" in command.upper(): 
                self.cursor.execute(command)
                return self.cursor.fetchall()
            else:
                self.cursor.execute(command)
                return []

    
    def create_table(self, name: str, columns: list):
        """
        Create Table on the previously set *.sqlite file
        
        Parameters:
            name    (str)   : The name of the table which cannot be duplicate and cannot be named "sqlite_sequence" as the name is being use for auto increment
            columns (list)  : Table columns name and datatypes in tuple which can be int, float, bool, str

        Returns:
            Table           : A Table SQL Class
        """
        self.__check_create_table(name=name, columns=columns)

        # Create SQL command
        sql = f"CREATE TABLE {name} (row_id INTEGER PRIMARY KEY AUTOINCREMENT,"
        for column in columns:
            sql += f"{column[0]} {SQLDATATYPE[column[1]]},"
        sql = sql[0:-1] + ");"
        
   
        with self.connection:
            self.cursor.execute(sql)

        # Return Table object
        return Table(name=name, connection=self.connection)

    def delete_table(self, name: str, keep_table: bool=False):
        """
        Delete table on database

        Parameters:
            name        (str)   : Name of the Table that you wish to be delete
            keep_table  (bool)  : Delete only data of the table
        """
        if keep_table:
            self.sqlcmd(f"TRUNCATE TABLE {name};")
        else:
            self.sqlcmd(f"DROP TABLE {name};")

    def get_table(self, name: str):
        """
        Create table objects from already exists table on database

        Parameters:
            name    (str) : Name of already exist table

        Returns:
            Table         : class table object
        """
        return Table(name=name, connection=self.connection)


    def __check_create_table(self, name: str, columns: list):
        invalids = [column for column in columns if not any([type(column) is tuple, len(column)==2, type(column[0]) in [str, int], column[0] != "row_id", column[1] in list(SQLDATATYPE.keys())])]
        if len(invalids) != 0 or name=="sqlite_sequence":
            raise CreateTableArgumentsError()








class CreateTableArgumentsError(Exception):
    """
    Exception raised when create table parameters are in an incorrect configurations
    """

    def __init__(self):
        self.message = "name or columns are in an incorrect configurations"
        super().__init__(self.message)

    def __str__(self):
        return f"{self.message}"

class InsertdatakeyMismatchError(Exception):
    """
    Exception raised when insert data dict key mismatch expected data
    """

    def __init__(self, rev, exp):
        self.message = f"Expected {exp} got {rev} instead"
    def __str__(self):
        return self.message
        



import sqlite3

from corgidb.constants import *
from corgidb.exceptions import *
from corgidb.objects.condition import Condition

class Table:
    """
    This is a class for manipulating SQL table can be generated from utils.create_table or utils.get_table method
    
    Attributes:
        name (str)  : The name of table
        connection (sqlite3.Connection) : Connection to Database

    """

    def __init__(self, name: str, connection: sqlite3.Connection):
        """
        The constructor for Table class

        Parameters:
            name (str)  : The name of table
            connection (sqlite3.Connection) : Connection to Database
        """
        self.name       = name
        self.connection = connection
        self.cursor     = self.connection.cursor()
    
    def get(self, condition: Condition):
        """
        This function query data from a table

        Attributes:
            condition (Condition) : Condition of the data that will be selected

        Return:
            dict    : dataframe like dictionary
        """
        

        cols = list(self.cols().keys())
        data = {c: [] for c in cols}
        print(data)
        
        if condition.sqlout != "":
            for row in self.__sqlcmd(f"SELECT * FROM {self.name} WHERE {condition.sqlout}"):
                for i, d in enumerate(row[1:]):
                    data[cols[i]].append(d)
        else:
            for row in self.__sqlcmd(f"SELECT * FROM {self.name}"):
                for i, d in enumerate(row[1:]):
                    data[cols[i]].append(d)
        
        return data

    def insert(self, data: dict):
        """
        This function insert new data to table function will convert dtype if possible

        Parameters:
            data (dict) : inserted data {col_name: data} 
        """
        cols       = self.cols()
        table_cols = set(cols.keys())
        
        # DataKey mismatch
        if set(data.keys())!=table_cols:
            raise InsertdatakeyMismatchError(set(data.keys), table_cols)

        table_cols = tuple(table_cols)
        insert_data = tuple([cols[col](data[col]) for col in table_cols])

        self.__sqlcmd(f"INSERT INTO {self.name} {table_cols} VALUES {insert_data};")
    
    def update(self, data:dict, condition: Condition):
        """
        This function update information on Table

        Parameters:
            data (dict) : data to be updated
            condition (Condition) : condition created using corgidb.sqlobjects to applied which row to be update
        """
        cols       = self.cols()
        table_cols = set(cols.keys())
        
        # DataKey mismatch
        if set(data.keys())!=table_cols:
            raise InsertdatakeyMismatchError(set(data.keys), table_cols)

        table_cols = tuple(table_cols)
        update_data = tuple([cols[col](data[col]) for col in table_cols])

        sql = ""
        for i,c in enumerate(table_cols):
            sql += f"{c} = {update_data[i]},"
        sql = sql[0:-1]

        self.__sqlcmd(f"UPDATE {self.name} SET {sql} WHERE {condition.sqlout};")

    def remove(self, condition: Condition):
        """
        This function delete information on Table

        Parameters:
            condition (Condition) : Condition created using corgidb.sqlobject to select row to be delete
        """
        self.__sqlcmd(f"DELETE FROM {self.name} WHERE {condition.sqlout};")

    def cols(self, dtype: str='Python', withid: bool=False) -> dict:
        """
        This function get the informations of Table Columns

        Parameters:
            dtype   (str) : setting for which dtype to be output Python(type default) or SQL(string)
            withid  (bool): Including row_id
        Returns:
            dict    : dictionary of columns {colname: dtypes} dtypes as python's type
        """
        dtype_inv = {v: k for k, v in SQLDATATYPE.items()}
        out = {}
        columns = self.__sqlcmd(f"PRAGMA table_info({self.name})").fetchall()

        if withid:
            for column in columns:
                if dtype.upper()=="SQL":
                    out[column[1]] = column[2]
                    continue
                if dtype.upper()=="PYTHON":
                    out[column[1]] = dtype_inv[column[2].upper()]
        else:
            for column in columns[1:]:
                if dtype.upper()=="SQL":
                    out[column[1]] = column[2]
                    continue
                if dtype.upper()=="PYTHON":
                    out[column[1]] = dtype_inv[column[2].upper()]
        return out

    def delete(self, keep_table:bool=False):
        """
        Delete table on database

        Parameters
            name        (str)   : Name of the Table that you wish to be delete
            keep_table  (bool)  : Delete only data of the table if this is False the object will be deleted as well
        """
        if keep_table:
            self.sqlcmd(f"TRUNCATE TABLE {self.name};")
        else:
            self.sqlcmd(f"DROP TABLE {self.name};")

    def __sqlcmd(self, command: str, raw=True):
        """
        This function to execute raw SQL command not recommened to be called if not necessary cannot return any data

        Parameters:
            command (str)   : Raw SQLite command
        
        Returns:
            list            : a list of fetched sql data return empty list if not fetched
        """
        if not raw:
            with self.connection:
                if "SELECT" in command.upper(): 
                    self.cursor.execute(command)
                    return self.cursor.fetchall()
                else:
                    self.cursor.execute(command)
                    return []
        else:
            with self.connection:
                return self.cursor.execute(command)





import sqlite3
from corgidb.objects.table import Table
from corgidb.utils import Utils
class CorgiDB:
    """
    Starting point of using CorgiDB as a sqlite wrapper

    Attributes:
        database_path (str): The path to database file *.sqlite
    """
    def __init__(self, database_path: str):
        self.connection = sqlite3.connect(database_path)
        self.cursor     = self.connection.cursor()

        # Initialize utils class
        self.utils      = Utils(connection=self.connection)




class Condition:
    """
    This is a class for creating SQLite Condition from python condition
    
    Attributes:
        conditions (list) : list of condition in lambda expression
        logic (str)     : logic to be applied between condition "AND" "OR"
    """

    def __init__(self, conditions: list, logic: str):
        """
        The constructor for Condition class
        
        Attributes:
            conditions (list) : list of condition in string ex. "age >= 5"
            logic (str)     : logic to be applied between condition "AND" "OR"
        """
        self.__conditions = conditions
        self.__logic      = logic

        if len(conditions) == 1:
            self.__logic=""
        
        if conditions!=[]:
            self.sqlout    = self.__sqlize()
        if conditions==[]:
            self.sqlout    = ""


    def __sqlize(self):
        sql = "("
        for condition in self.__conditions:
            if type(condition) == Condition:
                condition = condition.sqlout    
            sql += condition
            sql += " "
            sql += self.__logic
            sql += ","
        sql = sql[0: -1]
        sql += ")"
        return sql
    



    

    



from setuptools import setup, Extension

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


setup(
  name = 'corgidb',         # How you named your package folder (MyLib)
  packages = ['corgidb', 'corgidb.objects'],   # Chose the same as "name"
  version = '0.1.13',      # Start with a small number and increase it with every change you make
  license='MIT',        # Chose a license from here: https://help.github.com/articles/licensing-a-repository
  long_description=long_description,
  long_description_content_type="text/markdown",
  description = 'Basic sqlite3 wrapper write as short as a corgi legs and as fast as a corgi',   # Give a short description about your library
  author = 'Wasin Silakong',                   # Type in your name
  author_email = 'wasin.silakong@gmail.com',      # Type in your E-Mail
  url = 'https://github.com/WasinUddy/CorgiDB',   # Provide either the link to your github or to your website
  download_url = 'https://github.com/WasinUddy/CorgiDB/archive/refs/tags/v0.1.13-alpha.tar.gz',    # I explain this later on
  keywords = ['sql python', 'python database', 'SQL', 'python sqlite'],   # Keywords that define your package best
  install_requires=[],
  classifiers=[
    'Development Status :: 3 - Alpha',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
    'Intended Audience :: Developers',      # Define that your audience are developers
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',   # Again, pick a license
    'Programming Language :: Python :: 3',      #Specify which pyhton versions that you want to support
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
  ])


from corgidb.corgidb import CorgiDB
from corgidb.utils import Utils
from corgidb.objects.condition import Condition
from corgidb.objects.datachunk import DataChunk
from corgidb.objects.table import Table